import javax.swing.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.geom.*;
import java.util.*;

/**
 * <b>Programm:</b> WaveThread<br>
 * <b>Copyright:</b> 2002 Andreas Gohr, Frank Schubert, Milan Altenburg<br>
 * <b>License:</b> GPL2 or higher<br>
 * <br>
 * <b>Info:</b> Class CalcThread, calculates Raster, holds array of Rasters
 */
public class CalcThread extends Thread
{
  static final int MAX_QUEUE_SIZE = 10;
  
  static final int RED_MASK = 0xff0000;
  static final int BLUE_MASK = 0x0000ff;
  static final int GREEN_MASK = 0x00ff00;
  static final int NUMBER_OF_BITS = 24;
  
  int _width;     // current width
  int _height;     // current height
  float _time;   // seconds
  boolean _run;  // for main while() of thread
  boolean _calculate_with_java;
  boolean _calculate_with_sinus;
  boolean _calculate_with_pointers;
  boolean _calculate_with_copy;

  // fifo "pointers" :)
  int _first;
  int _last;
  int _length;

  int[] _vals; // array for calculated values

  MonoWave[] _mw;
  Color _color;
  DirectColorModel _dcm;
  WritableRaster[] _rasterqueue = new WritableRaster[MAX_QUEUE_SIZE];

  /**
   * CalcThread Constructor
   * sets time, MonoWave and Color
   * creates new RGB ColorModel
   */
  public CalcThread(MonoWave[] mw,Color col){
    _time  = 0;
    _mw = mw;
    _color = col;
    _calculate_with_java = true;
    _calculate_with_sinus = true;
    _calculate_with_pointers = false;
    _calculate_with_copy = false;
    System.loadLibrary("CalcThread");

    // bits, bitmasks for r,g & b
    _dcm = new DirectColorModel(NUMBER_OF_BITS,RED_MASK,GREEN_MASK,BLUE_MASK);
  }

  /**
   * sets new width and height
   * starts Thread, creates suitable Raster
   * resets queue values (length,first,last)
   */
  public synchronized void setXY(int x, int y){
    _width = x;
    _height = y;
    _run = true;
    _length=0;
    _vals = new int[3*_width*_height];

    // create suitable Rasters
    for (int i=0; i < MAX_QUEUE_SIZE; i++){
      _rasterqueue[i] = _dcm.createCompatibleWritableRaster(_width,_height);
    }

    // if first run call start()
    if(!this.isAlive()){
      this.start();
    }
  }

  /**
   * @return size of queue
   */
  public int getQueueSize(){
    return(_length);
  }

  /**
   * @return the Raster, that is next
   */
  public WritableRaster getRaster() {
    int tmp_first = _first; // hold value for return
    _first++;
    if(_first >= MAX_QUEUE_SIZE) _first = 0;
    _length--;
    if (_first == _last) _length=0;

    return(_rasterqueue[tmp_first]);
  }

  /**
   * control thread from outside, supposed but never used
   */
  public void die() {
    _run = false;
  }

  /**
	 * switches usage of either native c or java calculation
	 */
  public void switchCalcMethod() {
    if(_calculate_with_java) { _calculate_with_java = false; }
    else { _calculate_with_java = true; }
  }

  /**
	 * switches usage of either sinus or taylor approximation
	 */
  public void switchSinusCalcMethod() {
    if(_calculate_with_sinus) { _calculate_with_sinus = false; }
    else { _calculate_with_sinus = true; }
  }

	/**
	 * switches usage of pointers in native c function
	 */
  public void switchUsageOfPointers() {
    if(_calculate_with_pointers) { _calculate_with_pointers = false; }
    else { _calculate_with_pointers = true; }
  }

	/**
	 * switches usage of line copy
	 */
  public void switchUsageOfCopy() {
    if(_calculate_with_copy) { _calculate_with_copy = false; }
    else { _calculate_with_copy = true; }
  }

	/**
	 * @return String with options currently set
	 */
  public String getOptions(){
    String options= new String();
    if(_calculate_with_java) {
      options = "Java\t";
    }else{
      options = "C   \t";
    }
    if(_calculate_with_sinus){
      options += "sin\t";
    }else{
      options += "   \t";
    }
    if(_calculate_with_pointers) {
      options += "pointer\t";
    }else{
      options += "       \t";
    }
		if(_calculate_with_copy) {
      options += "copy\t";
    }else{
      options += "    \t";
    }

    return(options);
  }

	/**
	 * @return factorial
	 */
  public long fak(int x) {
    int i;
    long fak = 1;
    for(i=1;i<=x;i++){
      fak *= i;
    }
    return(fak);
  }

  /**
	 * @return taylor approximation of sinus
	 */
  public float sinusApproximation(float akt){
    boolean negate = false;

    if(akt > Math.PI) {
      akt -= (float)Math.PI;
      negate = true;
    }
    akt = (akt - (float)Math.pow(akt,3.0)/(float)fak(3)
               + (float)Math.pow(akt,5.0)/(float)fak(5)
               - (float)Math.pow(akt,7.0)/(float)fak(7)
               + (float)Math.pow(akt,9.0)/(float)fak(9)
               - (float)Math.pow(akt,11.0)/(float)fak(11)
               + (float)Math.pow(akt,13.0)/(float)fak(13)
               - (float)Math.pow(akt,15.0)/(float)fak(15)
               + (float)Math.pow(akt,17.0)/(float)fak(17));

     if(negate) { akt *= -1.0f; }
     return(akt);
  }

 /**
   * calculates the next Raster, according to time _time
   * @param index, where to put Raster
   */
  public void calc_java(int idx){
    int base, count=0, lines_timeo_calculate;
    float akt,value;
		long color;
    _vals = new int[3*_width*_height];

    for(int i=0;i<_mw.length;i++) {
      if(_mw[i].is_enabled()){
        count++;
      }
    }

		// if copy mode, only calculate one line
		if(_calculate_with_copy){
	    lines_timeo_calculate = 1;
	  }else{
	    lines_timeo_calculate = _height;
	  }

		//for each line
    for(int row=0;row<lines_timeo_calculate;row++) {
			for (int i=0; i<_width; i++){
				value = 0;
    		for(int wavenum=0; wavenum<count; wavenum++){
      		// if wave enabled
      		if(_mw[wavenum].is_enabled()) {

          	akt = (_mw[wavenum].getWaveKey() * i/_mw[wavenum].getScale()) -
                	(_mw[wavenum].getCircleFrequency() * 
                				_time/_mw[wavenum].getScale()) + 
                						_mw[wavenum].getPhase();

          	akt = akt%((float)(2.0*Math.PI));

          	if(_calculate_with_sinus) {
            	akt = (float) Math.sin((double) akt);
          	}else {
            	akt = sinusApproximation(akt);
          	}

          	akt = (akt+1)/2;
            akt *= _mw[wavenum].getRatio();

            value += akt;
				  }
				}

				color = (long) (2 * 255 * (value/count) );
        base = (3 * i) + (row * _width * 3);

				if(color < 255){
        	_vals[base+0]= 0;
        	_vals[base+1]= (int) (color);
        	_vals[base+2]= 0;
      	}else{
        	_vals[base+0]= 0;
        	_vals[base+1]= (int) (255);
        	_vals[base+2]= (int) (color-255);
      	}
    	}
	  }


		if(_calculate_with_copy) {
    	// fill other rows, copy
    	for(int j=1; j<_height; j++) {
      	for(int i=0;i<_width; i++) {
        	base = 3*j*_width + 3 * i;

        	_vals[base+0]= _vals[3*i+0];
        	_vals[base+1]= _vals[3*i+1];
        	_vals[base+2]= _vals[3*i+2];
      	}
    	}
    }

    (_rasterqueue[idx]).setPixels(0,0,_width,_height,_vals);
  }

  /**
   * declaration of native function for calculation of the raster
   */
  public native void calcArray(int[] returnvals,
                               float[] wavevals,
                               int x,
                               int y,
                               float t,
                               boolean sinus_method,
                               boolean use_pointers,
                               boolean use_copy);

  /**
   * calculates the next Raster, according to time _time but using a native function
   * @param index, where to put Raster
   */
  public void calc_native(int idx){
    int count=0;
    _vals = new int[3*_width*_height];

    for(int i=0;i<_mw.length;i++) {
      if(_mw[i].is_enabled()){
        count++;
      }
    }
    float[] waves = new float[count * 5];
    int j=0;
    for (int i=0; i< _mw.length; i++){
      if (_mw[i] != null && _mw[i].is_enabled()){
        waves[j*5]   = _mw[i].getWaveKey();
        waves[j*5+1] = _mw[i].getCircleFrequency();
        waves[j*5+2] = _mw[i].getPhase();
        waves[j*5+3] = _mw[i].getScale();
        waves[j*5+4] = _mw[i].getRatio(); //Amplitude between 0.0 and 1.0
        j++;
      }
    }

	// call native function
    calcArray(_vals,waves,_width,_height,_time,_calculate_with_sinus,
    						_calculate_with_pointers,_calculate_with_copy);
    
    (_rasterqueue[idx]).setPixels(0,0,_width,_height,_vals);
  }


  /**
   * "main" of CalcThread
   * calls synchronized calc to fill queue
   */
  public void run(){
    while( _run ) {
      if(_length==0){
        _first=0;
        _last=0;
        _time += 0.2;
        // avoid setXY() when calculating raster!
        synchronized(this){
          if(_calculate_with_java) { calc_java(0); }
          else { calc_native(0); }
          this.notify();
        }
        _length++;
      }else if(_length<MAX_QUEUE_SIZE-1 ){
        _last++;
        if(_last >= MAX_QUEUE_SIZE) _last=0;
        _time += 0.2;
        synchronized(this){
          if(_calculate_with_java) { calc_java(_last); }
          else { calc_native(_last); }
          this.notify();
        }
        _length++;
      }else{
        yield();
      }
    }
  }
}
